#ifdef HAVE_CONFIG_H
#include "config.h"
#endif
#include <norm.h>

#include <string.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <stdio.h>
#include <stdlib.h>
#include <net/ethernet.h>
#ifdef LINUX
#define __FAVOR_BSD
#endif
#include <netinet/in_systm.h>
#include <netinet/in.h>
#include <netinet/ip.h>
#include <netinet/udp.h>
#include <netinet/tcp.h>
#include <netinet/ip_icmp.h>

#ifdef FREEBSD
#include <netinet/ip_var.h>
#include <netinet/udp_var.h>
#endif
#include <sys/time.h>
#include <errno.h>
#include <netdb.h>
#include <pcap.h>


#define FDDI_HDR_LEN    13
#define SNAP_LEN        8

static char             errbuf[PCAP_ERRBUF_SIZE];
static pcap_t         *pcap;
static char           *input;
static char           *output;
static pcap_dumper_t  *dumper;   

static int             fragsize;
struct in_addr         ipaddr;
struct in_addr         mask;
int                    maskbits;

static u_short
in_cksum(register u_short *addr, register int len, u_int preadd)
{
  register int nleft = len;
  register u_short *w = addr;
  register u_short answer;
  register int sum = 0;

  sum += preadd;

  /*
   *  Our algorithm is simple, using a 32 bit accumulator (sum),
   *  we add sequential 16 bit words to it, and at the end, fold
   *  back all the carry bits from the top 16 bits into the lower
   *  16 bits.
   */
  while (nleft > 1) {
    sum += *w++;
    nleft -= 2;
  }

  /* mop up an odd byte, if necessary */
  if (nleft == 1)
    sum += *(u_char *)w;

  /*
   * add back carry outs from top 16 bits to low 16 bits
   */
  sum = (sum >> 16) + (sum & 0xffff); /* add hi 16 to low 16 */
  sum += (sum >> 16);                 /* add carry */
  answer = ~sum;                      /* truncate to 16 bits */

  return answer;
}


static u_int16_t 
ip_checksum(struct ip* iphdr)
{
  u_int16_t    old, result = 0;

  if (!iphdr)
    return (0);
  
  old = iphdr->ip_sum;
  iphdr->ip_sum = 0;
  result = in_cksum((u_short *)iphdr, iphdr->ip_hl<<2, 0);
  iphdr->ip_sum = old;

  return result;
}


static void
handler(u_char *user, const struct pcap_pkthdr* const_pkthdr, const u_char * packet)
{
  struct pcap_pkthdr  *pkthdr = (struct pcap_pkthdr*)const_pkthdr;
  struct ip           *iphdr  = NULL;
  u_char              *offset = NULL;
  u_char              *payload = NULL;
  int                  todo, ip_hl, link_len, frag_off;
  u_int16_t            ip_off;
  
  switch (pcap_datalink(pcap))
    {
    case DLT_EN10MB:
      {
        struct ether_header *eh;
	struct snap_header  *sh;

	eh = (struct ether_header*)packet;
	
	/* Is this an Ethernet-II frame or a Ethernet 802.3 SNAP frame?*/
	/* Check ether_type > 1500 --> Ethernet II */
	if (htons(eh->ether_type) <= ETHERMTU)
	  {
	    sh = (struct snap_header*)(packet + ETHER_HDR_LEN);
	    
	    if (sh->ssap != 0xaa)
	        return;
	    
	    /* Check for IP payload: 0x00-00-00-08-00 SNAP field */
	    if (((sh->snap[0]) | (sh->snap[1]) | (sh->snap[2]) | 
		 (!sh->snap[4])) && (sh->snap[3]==8))
	      offset = (u_char*) (packet + (ETHER_HDR_LEN + SNAP_LEN));
	  }
	else 
	  {
	    if (htons(eh->ether_type) == ETHERTYPE_IP)
	      offset = (u_char*) (packet + (ETHER_HDR_LEN));
	    else
	      return;
	  }
      }
      break;

    case DLT_FDDI:
      {
	struct fddi_header *fh;
	struct snap_header *sh;

	fh = (struct fddi_header*)packet;
	sh = (struct snap_header*)(packet + FDDI_HDR_LEN);


	/*Is this an FDDI 802.2 or an FDDI SNAP frame?*/
	if ((sh->ssap==0xaa)&&(sh->dsap==0xaa))
	  {
	    /*It's a SNAP frame*/
	    if (((sh->snap[0]) | (sh->snap[1]) | (sh->snap[2]) |
		 (sh->snap[4]==0)) &&(sh->snap[3]==8))
	      /*It's a SNAP header containing an IP datagram*/
	      offset = (u_char*) (packet + (FDDI_HDR_LEN + SNAP_LEN));
	    else
	      return;
	  }
	else
	  return;
      }
      break;

    default:
      offset = (u_char*)packet;
    }

  if (!offset)
    return;

  /* Offset is at IP hdr now: */
  iphdr = (struct ip*)offset;

  ip_hl = iphdr->ip_hl * 4;
  ip_off = ntohs(iphdr->ip_off);

  todo = ntohs(iphdr->ip_len) - ip_hl;
  offset = payload = (u_char*)iphdr + ip_hl;
  frag_off = 0;
  link_len = pkthdr->caplen - ntohs(iphdr->ip_len);

  if (todo < fragsize                       ||
      ((ntohl(iphdr->ip_src.s_addr) & mask.s_addr) - ipaddr.s_addr != 0)  ||
      pkthdr->caplen != pkthdr->len)
    {
      pcap_dump((u_char*)dumper, pkthdr, packet);
      return;
    }
  
  while (todo > 0)
    {
      ip_off = frag_off; 
      
      if (todo >= fragsize)
	{
	  memmove(payload, offset, fragsize);

	  if (todo > fragsize)
	    ip_off |= IP_MF;
	  
	  iphdr->ip_len = htons(ip_hl + fragsize);
	  pkthdr->caplen = link_len + ip_hl + fragsize;
	  pkthdr->len = link_len + ip_hl + fragsize;	  
	}
      else
	{
	  memmove(payload, offset, todo);

	  iphdr->ip_len = htons(ip_hl + todo);
	  pkthdr->caplen = link_len + ip_hl + todo;
	  pkthdr->len = link_len + ip_hl + todo;	  
	}

      iphdr->ip_off = htons(ip_off);
      iphdr->ip_sum = ip_checksum(iphdr);

      pcap_dump((u_char*)dumper, pkthdr, packet);

      offset += fragsize;
      todo -= fragsize;
      frag_off += (fragsize / 8);
    }
  
  return;
  user = NULL;
}


int
main(int argc, char **argv)
{
  if (argc < 6)
    {
      printf("USAGE: %s <fragsize in 8-byte words> <ipaddr> <netmaskbits> <input trace> <output trace>\n", argv[0]);
      exit(1);
    }

  fragsize = atoi(argv[1]) * 8;

  inet_aton(argv[2], &ipaddr);
  ipaddr.s_addr = ntohl(ipaddr.s_addr);

  maskbits = atoi(argv[3]);
  mask.s_addr = ~( (2 << (32-maskbits-1)) - 1);

  input    = argv[4];
  output   = argv[5];

  pcap = pcap_open_offline(input, errbuf);
  dumper = pcap_dump_open(pcap, output);
  pcap_loop(pcap, 0, handler, NULL);
  pcap_close(pcap);
  pcap_dump_close(dumper);


  return (0);
}
