#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <stdlib.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <net/if.h>
#include <sys/ioctl.h>
#include <unistd.h>
#include <errno.h>

#include <netinet/in_systm.h>
#include <netinet/in.h>
#include <netinet/ip.h>
#include <fcntl.h>
#include <string.h>
#include <arpa/inet.h>

#include <norm.h>
#include <norm_debug.h>
#include <norm_ether.h>
#include <norm_fddi.h>
#include <norm_ip_fragment.h>
#include <norm_packet.h>
#include <norm_options.h>
#include <norm_tcb.h>
#include <norm_send.h>
#include <norm_stats.h>


Interface    nifs[NUM_IFS];
int          sock = 0;


void
init_offline_interface(Interface *nif)
{
  char            ebuf[256];
  pcap_t         *pcap;

  D_ENTER;

  pcap = PCAP_OPEN_OFFLINE(options.in_tracefile, ebuf);
  if (!pcap)
    {
      perror("Failed to open file");
      exit(1);
    }
  
  nif->dlt = pcap_datalink(pcap);
  nif->s = open("/dev/null", O_WRONLY);
  if (nif->s < 0)
    {
      perror("Failed to open /dev/null!");
      exit(1);
    }
  
  if ( (nif->pcap_dumper = PCAP_DUMP_OPEN(pcap, options.out_tracefile)) == NULL)
    {
      perror("Failed to open dumpfile!");
      exit(1);
    }
  nif->pcap = pcap;
  nif->dev = NULL;
  D_RETURN;
}


void
init_online_interface(Interface *nif, int if_num)
{
  int                  one = 1;
  char                 errbuf[256];
  char                 hwaddr[20];
  char                 hwaddrstr[20];
  pcap_t              *pcap;
  struct ifreq         ifr;
  struct sockaddr_in  *addr;
  struct bpf_program   fp;
  char                 prog[256];

  D_ENTER;
  
  /* Set interface name: */
  nif->dev = strdup(options.ifs[if_num]);

  /* Find the MAC address of this interface */
  if (sock == 0)
    {
      sock = socket(AF_INET, SOCK_DGRAM, 0);
      if (sock < 0) {
	perror("socket failed");
	exit(1);
      }
    }

  strcpy(ifr.ifr_name, nif->dev);
  if (ioctl(sock, SIOCGIFADDR, &ifr)<0) {
    perror("ioctl SIOCGIFADDR failed");
    exit(1);
  }

  /* Fill in ASCII interface address */
  addr = (struct sockaddr_in*) &ifr.ifr_addr;
  nif->addrs = inet_ntoa(addr->sin_addr);
  norm_ether_get_addr(addr->sin_addr, hwaddr);

  D(NOISE_LOGLEVEL, ("Device %s has IP %s at %s\n", nif->dev,
		     nif->addrs, norm_ether_etoa(hwaddr)));

  pcap = pcap_open_live(nif->dev, 1600, 0, 1, errbuf);

  if (pcap == NULL)
    {
      fprintf(stderr, "norm:  %s\n", errbuf);
      D_RETURN;
    }

  /* Setup packet filter. We want a filter that accepts
     packets with our dst MAC address, but not our dst IP address.
  */
  sprintf(hwaddrstr, "%x:%x:%x:%x:%x:%x", hwaddr[0]&0xff, hwaddr[1]&0xff, 
	  hwaddr[2]&0xff, hwaddr[3]&0xff, hwaddr[4]&0xff, hwaddr[5]&0xff);
  sprintf(prog, "ether dst %s and not ether src %s and not dst %s", 
	  hwaddrstr, hwaddrstr, inet_ntoa(addr->sin_addr));
  D(NOISE_LOGLEVEL, ("compiling: %s\n", prog));
  if (pcap_compile(pcap, &fp, prog, 0, 0xffffff80)<0)
    pcap_perror(pcap, "error compiling");
  if (pcap_setfilter(pcap, &fp)<0)
    pcap_perror(pcap, "error setting filter");
  
  /* Fill in link type. */
  nif->dlt = pcap_datalink(pcap);

  /* Fill in IPv4 structure */
  bzero(&(nif->ifip), sizeof(struct sockaddr_in));
  nif->ifip.sin_family = AF_INET;
  nif->ifip.sin_port = 0;
  inet_aton(nif->addrs, &(nif->ifip.sin_addr));
  
  /* Generate raw socket for sending packets and connect. */
  nif->s = socket(PF_INET, SOCK_RAW, IPPROTO_RAW);

  if (connect(nif->s, (struct sockaddr *) &(nif->ifip), sizeof(struct sockaddr_in)) < 0)
    perror("connect:");
  if (setsockopt(nif->s, IPPROTO_IP, IP_HDRINCL, &one, sizeof(int)) < 0)
    perror("setsockopt HDRINCL:");

  nif->pcap = pcap;
  D_RETURN;
}


/* pcap Ethernet handler
*/
void pcap_callback_ether(u_char *user, const struct pcap_pkthdr* pkthdr, const u_char *pdata)
{
  NormPacket    p;
  Interface    *nif;
  int           num;

  D_ENTER;
  fflush(stdout);

  if (pkthdr->caplen != pkthdr->len)
    {
      printf("didn't capture enough data! caplen: %d pkthdr->len: %d\n",
	      pkthdr->caplen, pkthdr->len);
      D_RETURN;
    }

  nif = (Interface *)user;
  bzero(&p, sizeof(NormPacket));
  p.ph = (struct pcap_pkthdr*)pkthdr;
  p.length = pkthdr->len;
  p.link_data = (u_char*)pdata;

  num = norm_stats_increment_num_packets();

  if (options.debuglevel >= NORM_LOGLEVEL)
    printf("[%i]----------------------------------------------------------------\n",
	   num);
  
  if (options.show_ether)
    {
      print_ether(NORM_LOGLEVEL, ("received on if %s (Ethernet) %d bytes\n", nif->dev, pkthdr->len));
    }

  if (options.passive)
    norm_packet_output(&p, nif);
  else
    norm_ether_handler(&p, nif);

  D_RETURN;
}


/* pcap FDDI handler
*/
void pcap_callback_fddi(u_char *user, const struct pcap_pkthdr* pkthdr, const u_char *pdata)
{
  NormPacket    p;
  Interface    *nif;
  int           num;

  D_ENTER;
  fflush(stdout);

  if (pkthdr->caplen != pkthdr->len) {
    fprintf(stderr, "didn't capture enough data! caplen: %d pkthdr->len: %d\n",
	    pkthdr->caplen, pkthdr->len);
    D_RETURN;
  }
  
  nif=(Interface *)user;
  bzero(&p, sizeof(NormPacket));
  p.ph = (struct pcap_pkthdr*)pkthdr;
  p.length = pkthdr->len;
  p.link_data = (u_char*)pdata;
  
  num = norm_stats_increment_num_packets();

  if (options.debuglevel >= NORM_LOGLEVEL)
    printf("[%i]----------------------------------------------------------------\n",
	   num);
  
  if (options.show_ether)
    {
      print_fddi(NORM_LOGLEVEL, ("received on if %s (FDDI) %d bytes\n", nif->dev, pkthdr->len));
    }

  if (options.passive)
    norm_packet_output(&p, nif);
  else
    norm_fddi_handler(&p, nif);

  D_RETURN;
}


int
main(int argc, char **argv)
{
  int devices, i, status = 1;

  norm_options_read(argc, argv);
  norm_options_install_sighandler();

  D_ENTER;

  norm_stats_init();
  norm_ip_fragment_init();
  norm_tcb_init();

  if (options.in_tracefile)
    {
      init_offline_interface(&(nifs[0]));
      devices=1;
    }
  else
    {
      /* Setup pcap stuff */
      for (i = 0; i < options.num_ifs; i++)
	{
	  init_online_interface(&(nifs[i]), i);
	}
      devices = options.num_ifs;
    }

  norm_stats_start_clock();

  fopen("start_of_real_work", "r");

  while (status)
    {
      for (i=0; i<devices; i++)
	{
	  switch (nifs[i].dlt)
	    {
	    case DLT_EN10MB:
	      status = PCAP_DISPATCH(nifs[i].pcap, 0, pcap_callback_ether, 
				 (u_char *)&(nifs[i]));
	      break;
	    case DLT_FDDI:
	      status = PCAP_DISPATCH(nifs[i].pcap, 0, pcap_callback_fddi, 
				 (u_char *)&(nifs[i]));
	      break;
	    default:
	      fprintf(stderr, "Unknown data link type: %d\n", nifs[i].dlt);
	      exit(1);
	    }
	  if (status < 0)
	    {
	      fprintf(stderr, "norm: packet dispatch error\n");
	    }
	}
    }

  fopen("end_of_real_work", "r");

  /*
  printf("%i packets processed in %f seconds (user), %f seconds wall, %f pckts/s\n",
	 norm_stats_get_num_packets(), norm_stats_get_usertime(),
	 norm_stats_get_walltime(), norm_stats_get_pps());
  */

  //printf("%f\n", norm_stats_get_pps_calculated(100000));
  // printf("%f\n", norm_stats_get_pps());
  
  printf("Did %i norms, %i IP, %i TCP, %i UDP, %i ICMP.\n",
	 norm_stats_get_norms(), norm_stats_get_ip_norms(),
	 norm_stats_get_tcp_norms(), norm_stats_get_udp_norms(),
	 norm_stats_get_icmp_norms());

  for(i=0; i<devices; i++)
    {
      PCAP_CLOSE(nifs[i].pcap);
      if (nifs[i].pcap_dumper)
	PCAP_DUMP_CLOSE(nifs[i].pcap_dumper);
    }

  if (sock)
    close(sock);

  D_RETURN_(0);
}
