#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <stdlib.h>
#include <err.h>
#include <string.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <net/if.h>
#ifndef LINUX
#include <net/if_dl.h>
#endif
#include <sys/ioctl.h>
#include <unistd.h>
#include <errno.h>

#include <netinet/in_systm.h>
#include <netinet/in.h>
#include <netinet/ip.h>
#include <stdio.h>
#include <sys/time.h>

#include "norm_ether.h"
#include "norm_ip.h"
#include "norm_misc.h"
#include "norm_debug.h"


int 
norm_ether_handler(NormPacket *p, Interface *nif)
{
  struct ether_header *eh;
  struct snap_header  *sh;
  unsigned char buf[ETHER_HDR_LEN+SNAP_LEN];

  D_ENTER;

  /*things may be unaligned - copy the header to ensure alignment*/
  memcpy(buf, p->link_data, min(p->length, ETHER_HDR_LEN+SNAP_LEN));
  eh = (struct ether_header*)buf;

  /* Is this an Ethernet-II frame or a Ethernet 802.3 SNAP frame?*/
  /* Check ether_type > 1500 --> Ethernet II */
  if (htons(eh->ether_type) <= ETHERMTU)
    {
#if 0
      ether_print(eh->ether_shost);
      printf("->");
      ether_print(eh->ether_dhost);
      printf("\n");
#endif
      memcpy(buf, p->link_data+ETHER_HDR_LEN, 
	     min(p->length - ETHER_HDR_LEN, SNAP_LEN));
      sh = (struct snap_header*)buf;
#if 0
      printf("dsap:%x ssap:%x ctrl:%x snap:%x,%x,%x,%x,%x\n",
	     sh->dsap, sh->ssap, sh->ctrl,
	     sh->snap[0]&0xff,
	     sh->snap[1]&0xff,
	     sh->snap[2]&0xff,
	     sh->snap[3]&0xff,
	     sh->snap[4]&0xff);
#endif
      if (sh->ssap != 0xaa)
      {
	printf("It's not Ethernet-II and it's not SNAP - what is it?\n");
	D_RETURN_(-1);
      }

      /* Check for IP payload: 0x00-00-00-08-00 SNAP field */
      if (((sh->snap[0]) | (sh->snap[1]) | (sh->snap[2]) | (!sh->snap[4])) && (sh->snap[3]==8))
      {
	/* It's a SNAP header containing an IP datagram */
	print_ether(NORM_LOGLEVEL, ("Handling SNAP Ethernet frame\n"));

	p->network_data = p->link_data + (ETHER_HDR_LEN+SNAP_LEN);
	D_RETURN_(norm_ip_handler(p, nif));
      }
    }
  else 
    {
      if (htons(eh->ether_type) == ETHERTYPE_IP)
	{
	  /* It's Ethernet-II containing an IP datagram */
	  print_ether(NORM_LOGLEVEL, ("Handling Ethernet-II frame\n"));

	  p->network_data = p->link_data + (ETHER_HDR_LEN);
	  D_RETURN_(norm_ip_handler(p, nif));
	}
      else
	{
	  print_ether(NORM_LOGLEVEL, ("bad ethernet type field (%x)- not IP\n", htons(eh->ether_type)));
	  D_RETURN_(-1);
	}
    }
  
  D_RETURN_(0);
}


char *
norm_ether_etoa(u_char *cp)
{
  static char s[18];

  D_ENTER;
  sprintf(s, "%x:%x:%x:%x:%x:%x", cp[0], cp[1], cp[2], cp[3], cp[4], cp[5]);
  D_RETURN_((char*)&s);
}



int
norm_ether_get_addr(struct in_addr ipaddr, u_char *hwaddr)
{
  struct ifreq *ifr, *ifend, *ifp;
  u_int32_t ina, mask;
#ifndef LINUX
  struct sockaddr_dl *dla;
#endif
  struct ifreq ifreq;
  struct ifconf ifc;
  struct ifreq ifs[MAX_IFS];
  int s;
  
  D_ENTER;

  s = socket(AF_INET, SOCK_DGRAM, 0);
  if (s < 0)
    err(1, "socket");
  
  ifc.ifc_len = sizeof(ifs);
  ifc.ifc_req = ifs;
  if (ioctl(s, SIOCGIFCONF, &ifc) < 0) {
    warnx("ioctl(SIOCGIFCONF)");
    close(s);
    D_RETURN_(0);
  }
  
  /*
   * Scan through looking for an interface with an Internet
   * address on the same subnet as `ipaddr'.
   */
  ifend = (struct ifreq *) (ifc.ifc_buf + ifc.ifc_len);

  for (ifr = ifc.ifc_req; ifr < ifend; ) {
    if (ifr->ifr_addr.sa_family == AF_INET) {

      ina = ((struct sockaddr_in *) &ifr->ifr_addr)->sin_addr.s_addr;
      strncpy(ifreq.ifr_name, ifr->ifr_name, sizeof(ifreq.ifr_name));

      /*
       * Check that the interface is up,
       * and not point-to-point or loopback.
       */
      if (ioctl(s, SIOCGIFFLAGS, &ifreq) < 0)
	continue;

      if ((ifreq.ifr_flags &
	   (IFF_UP|IFF_BROADCAST|IFF_POINTOPOINT|
	    IFF_LOOPBACK|IFF_NOARP))
	  != (IFF_UP|IFF_BROADCAST))
	goto nextif;

      /*
       * Get its netmask and check that it's on 
       * the right subnet.
       */
      if (ioctl(s, SIOCGIFNETMASK, &ifreq) < 0)
	continue;

      mask = ((struct sockaddr_in *) &ifreq.ifr_addr)->sin_addr.s_addr;

      if ((ipaddr.s_addr & mask) != (ina & mask))
	goto nextif;

      break;
    }

  nextif:

#ifdef LINUX
    ifr++;
#else
    ifr = (struct ifreq *) ((char *)&ifr->ifr_addr + ifr->ifr_addr.sa_len);
#endif  
  }
  
  if (ifr >= ifend) {
    close(s);
    D_RETURN_(0);
  }
  
  /*
   * Now scan through again looking for a link-level address
   * for this interface.
   */
  ifp = ifr;

  for (ifr = ifc.ifc_req; ifr < ifend; ) {
#ifdef LINUX
    if (strcmp(ifp->ifr_name, ifr->ifr_name) == 0 && ifr->ifr_addr.sa_family == AF_NETLINK)
#else
    if (strcmp(ifp->ifr_name, ifr->ifr_name) == 0 && ifr->ifr_addr.sa_family == AF_LINK)
#endif
      {
	/*
	 * Found the link-level address - copy it out
	 */

#ifdef LINUX
#define HACK_HWADDRLEN  (IFHWADDRLEN)
#else
#define HACK_HWADDRLEN  (dla->sdl_alen)
#endif

#ifdef LINUX
	memcpy(hwaddr, &ifr->ifr_addr, HACK_HWADDRLEN);
	printf("using interface %s for proxy with address: ", ifp->ifr_name);
#else
	dla = (struct sockaddr_dl *) &ifr->ifr_addr;
	memcpy(hwaddr,  LLADDR(dla), HACK_HWADDRLEN);
#endif
	close (s);
	D_RETURN_(HACK_HWADDRLEN);
      }
#ifdef LINUX
    ifr++;
#else
    ifr = (struct ifreq *) ((char *)&ifr->ifr_addr + ifr->ifr_addr.sa_len);
#endif
  }
  
  D_RETURN_(0);
}

