/* Contains code derived from traceroute, which has the following copyright:
 *
 * Copyright (c) 1988, 1989, 1991, 1994, 1995, 1996, 1997
 *    The Regents of the University of California.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that: (1) source code distributions
 * retain the above copyright notice and this paragraph in its entirety, (2)
 * distributions including binary code include the above copyright notice and
 * this paragraph in its entirety in the documentation or other materials
 * provided with the distribution, and (3) all advertising materials mentioning
 * features or use of this software display the following acknowledgement:
 * ``This product includes software developed by the University of California,
 * Lawrence Berkeley Laboratory and its contributors.'' Neither the name of
 * the University nor the names of its contributors may be used to endorse
 * or promote products derived from this software without specific prior
 * written permission.
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND WITHOUT ANY EXPRESS OR IMPLIED
 * WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
 */


#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <stdlib.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <stdio.h>
#include <sys/time.h>
#include <errno.h>
#include <netinet/in.h>
#include <arpa/inet.h>

#include <norm.h>
#include <norm_debug.h>
#include <norm_icmp.h>
#include <norm_ip.h>
#include <norm_ip_fragment.h>
#include <norm_misc.h>
#include <norm_send.h>
#include <norm_tcp.h>
#include <norm_udp.h>
#include <norm_stats.h>


#define DROP_LSRR
#define DROP_SSRR
#define DROP_RR
#define DEFAULT_TTL   10

#ifndef IPPROTO_IGRP
#define IPPROTO_IGRP  88
#endif
#ifndef IPPROTO_PGM
#define IPPROTO_PGM  113
#endif


int     
norm_ip_handler(NormPacket *p, Interface *nif)
{
  struct ip      *iphdr;
  u_short         off;
  int             len;

  iphdr = (struct ip*)p->network_data;
  len = norm_packet_get_length_ip(p);

  D_ENTER;

  /*sanity check: do we have enough data here to be a packet?*/
  if ((u_int)len < sizeof(struct ip)) {
    print_ip(NORM_LOGLEVEL, ("packet is too small to be IP -- dropping.\n"));
    D_RETURN_(-1);
  }
  
  /*sanity check: is it IP?*/
  if (iphdr->ip_v != 4)
    {
      print_ip(NORM_LOGLEVEL, ("packet is not IPv4! ip_v=%u -- dropping.\n", iphdr->ip_v));
      D_RETURN_(-1);
    }

  /*sanity check: is total length correct?*/
  if (options.ip_total_length_check)
    {
      if (ntohs(iphdr->ip_len) < len)
	{
	  int delta;

	  delta = (ntohs(iphdr->ip_len) - len);

	  print_ip(NORM_LOGLEVEL, ("Removing %i bytes to match ip_len (%u).\n",
				   (len - ntohs(iphdr->ip_len)), ntohs(iphdr->ip_len)));
	  p->length += delta;
	  p->ph->caplen += delta;
	  p->ph->len += delta;
	  len = ntohs(iphdr->ip_len);
	  norm_stats_increment_ip_norms();
	}

      if (ntohs(iphdr->ip_len) != len)
	{
	  norm_stats_increment_ip_norms();
	  print_ip(NORM_LOGLEVEL, ("total length incorrect; is %d should be %d, dropping.\n",
				   ntohs(iphdr->ip_len), len));
	  D_RETURN_(-1);
	}
    }
  
  /* sanity check: is header length field sane? */
  if (options.ip_header_length_check)
    {
      if (iphdr->ip_hl < 5)
	{
	  norm_stats_increment_ip_norms();
	  print_ip(NORM_LOGLEVEL, ("header length too small (%d) -- dropping.\n", iphdr->ip_hl));
	  D_RETURN_(-1);
	}
      if (iphdr->ip_hl*4 > len)
	{
	  norm_stats_increment_ip_norms();
	  print_ip(NORM_LOGLEVEL, ("header length too large: %d packet length is %d, dropping.\n",
				   iphdr->ip_hl, len));
	  D_RETURN_(-1);
	}
    }
  
  /* sanity check: is the header checksum OK? */
  if (options.ip_header_checksum_check)
    {
      if (norm_ip_checksum_valid(iphdr) == 0)
	{
	  norm_stats_increment_ip_norms();
	  print_ip(NORM_LOGLEVEL, ("Invalid IP checksum -- dropping.\n"));
	  D_RETURN_(-1);
	}
    }

  /* clear TOS flags if desired */
  if (options.ip_header_clear_tos)
    {
      if (iphdr->ip_tos != 0)
	{
	  norm_stats_increment_ip_norms();
	  print_ip(NORM_LOGLEVEL, ("Clearing TOS field.\n"));
	  iphdr->ip_tos = 0;
	  p->network_data_recalc_checksum = TRUE;
	}
    }

  
  if (options.ip_clear_rf && (ntohs(iphdr->ip_off) & IP_RF))
    {
      norm_stats_increment_ip_norms();
      print_ip(NORM_LOGLEVEL, ("Clearing reserved fragment flag.\n"));
      iphdr->ip_off = htons(ntohs(iphdr->ip_off) & ~IP_RF);
    }

  if (options.ip_header_addr_check)
    {
      /*is the source address sane?*/
      {
	unsigned int topbyte;
	topbyte = (ntohl(iphdr->ip_src.s_addr) & 0xff000000)>>24;
	if ((topbyte>223)||(topbyte==127)||(topbyte==0))
	  {
	    norm_stats_increment_ip_norms();
	    print_ip(NORM_LOGLEVEL, ("invalid source address: %s, dropping.\n",
				     inet_ntoa(iphdr->ip_src)));
	    D_RETURN_(-1);
	  }
      }

      /*is the destination address sane?*/
      {
	unsigned int topbyte;
	topbyte = (ntohl(iphdr->ip_dst.s_addr) & 0xff000000)>>24;
	/*allow multicast, but not class E or above*/
	if ((topbyte>239)||(topbyte==127)||(topbyte==0))
	  {
	    norm_stats_increment_ip_norms();
	    print_ip(NORM_LOGLEVEL, ("invalid destination address: %s, dropping.\n",
				     inet_ntoa(iphdr->ip_dst)));
	    D_RETURN_(-1);
	  }
      }
    }
  
  off = ntohs(iphdr->ip_off);

  /*
  {

    strcpy(srcstr, inet_ntoa(iphdr->ip_src));
    strcpy(dststr, inet_ntoa(iphdr->ip_dst));
    
    print_ip(INFO_LOGLEVEL, ("%s->%s src:%s dst:%s len:%d DF: %d\n",
			     (norm_misc_is_trusted_address(iphdr->ip_src) ? "trusted" : "untrusted"),
			     (norm_misc_is_trusted_address(iphdr->ip_dst) ? "trusted" : "untrusted"),
			     srcstr, dststr, len, (off & IP_DF)!=0));
  }
  */

  /* is this a fragment? */
  if (((off & IP_MF) != 0) || (off & IP_OFFMASK) != 0)
    {
      print_ip(INFO_LOGLEVEL, ("fragment --  offset(%d)  DF(%c)  MF(%c)\n",
			       ((off & IP_OFFMASK) << 3),
			       (off & IP_DF ? '*' : ' '),
			       (off & IP_MF ? '*' : ' ')));
      
      /* drop fragments when offset + packet length > 65535. */
      if (options.ip_check_fragment_offset &&
	  ((off & IP_OFFMASK) << 3) + len > IP_MAXPACKET)
	{
	  norm_stats_increment_ip_norms();
	  print_ip(NORM_LOGLEVEL,
		   ("Fragment offset + packet length > IP_MAXPACKET -- dropping.\n"));
	  D_RETURN_(-1);
	}
      
      /* drop fragments with DF flag if desired. */
      if (options.ip_drop_fragments_with_df && (off & IP_DF))
	{
	  norm_stats_increment_ip_norms();
	  print_ip(NORM_LOGLEVEL, ("DF set on fragment -- dropping.\n"));
	  D_RETURN_(-1);
	}
      
      D_RETURN_(norm_ip_fragment_handler(p, nif) < 0);
      
      /* IP fragment's payload is checked only after fragments
	 are reassembled ... */
    }

  D_RETURN_(norm_ip_handle_regular_packet(p, nif));
}



int
norm_ip_handle_regular_packet(NormPacket *p, Interface *nif)
{
  D_ENTER;

  if (options.ip_check_options)
    {
      if (norm_ip_process_options(p, nif) < 0)
	D_RETURN_(-1);
    }

  norm_ip_apply_normalizations(p, nif);

  if (norm_ip_process_protocol(p, nif) < 0)
    D_RETURN_(-1);

  /* If we've touched the packet, fix checksum: */
  if (p->network_data_recalc_checksum)
    {
      struct ip      *iphdr;

      iphdr = (struct ip*)p->network_data;
      iphdr->ip_sum = 0;
      iphdr->ip_sum = norm_ip_in_cksum((u_short *) iphdr, iphdr->ip_hl<<2, 0);      
      print_ip(INFO_LOGLEVEL, ("Recalculating IP checksum, now: %i\n", ntohs(iphdr->ip_sum)));
    }
  
  norm_packet_send(p, nif);
  D_RETURN_(0);
}


int
norm_ip_process_options(NormPacket *p, Interface *nif)
{
  u_char    *optp;
  int        optslen = 0;
  int        optsdone = 0;
  int        olen;
  struct ip *iphdr;

  D_ENTER;

  iphdr = (struct ip*) p->network_data;
  optslen = (iphdr->ip_hl)*4 - 20;
  optp = p->network_data + 20;

  while (optsdone < optslen)
    {
      switch(*optp)
	{
	case 0:
	  /*end of option list - RFC791*/
	  optsdone = optslen;
	  break;
	case 1:
	  /* No op*/
	  optp++;
	  optsdone++;
	  break;
	case 130:
	case 133:
	  /*Security - see RFC1108*/
	  /*we sanity check this, but otherwise pass it normally*/
	  optp++;
	  olen=*optp;
	  optp+=olen-1;
	  optsdone+=olen;
	  break;
	case 131:
	  /*Loose Source and Record Route - RFC791*/
#ifdef DROP_LSRR
	  print_ip(NORM_LOGLEVEL, ("dropping packet with LSRR option\n"));
	  D_RETURN_(-1);
#else
	  optp++;
	  olen=*optp;
	  print_ip(INFO_LOGLEVEL, ("LSRR olen: %d\n", olen));
	  optp+=olen-1;
	  optsdone+=olen;
	  break;
#endif    
	case 137:
	  /*Strict Source and Record Route - RFC791*/
#ifdef DROP_SSRR
	  print_ip(NORM_LOGLEVEL, ("dropping packet with SSRR option\n"));
	  D_RETURN_(-1);
#else
	  optp++;
	  olen=*optp;
	  optp+=olen-1;
	  optsdone+=olen;
	  break;
#endif    
	case 7:
	  /*Record Route - RFC791*/
#ifdef DROP_RR
	  print_ip(NORM_LOGLEVEL, ("dropping packet with RR option\n"));
	  D_RETURN_(-1);
#else
	  optp++;
	  olen=*optp;
	  optp+=olen-1;
	  optsdone+=olen;
	  break;
#endif    
	case 136:
	  /*Stream ID - RFC791*/
	  /*we sanity check this, but otherwise pass it normally*/
	  optp++;
	  olen=*optp;
	  if (olen!=4) {
	    print_ip(NORM_LOGLEVEL, ("Incorrect stream ID length: %d -- dropping.\n", olen));
	    D_RETURN_(-1);
	  }
	  optp+=olen-1;
	  optsdone+=olen;
	  break;
	case 68:
	  /*Internet timestamp - RFC791*/
	  /*harmless...*/
	  optp++;
	  olen=*optp;
	  optp+=olen-1;
	  optsdone+=olen;
	  break;
	case 148:
	  /*Router Alert - See RFC2113*/
	  /*we sanity check this, but otherwise pass it normally*/
	  optp++;
	  olen=*optp;
	  if (olen!=4) {
	    print_ip(NORM_LOGLEVEL, ("Incorrect router alert length: %d -- dropping.\n", olen));
	    D_RETURN_(-1);
	  }
	  optp+=olen-1;
	  optsdone+=olen;
	  break;
	default:
	  print_ip(NORM_LOGLEVEL, ("unknown option type: %d\n", *optp));
	  D_RETURN_(-1);
	}

      if (optsdone>optslen)
	{
	  print_ip(NORM_LOGLEVEL, ("bogus variable-length option - dropping.\n"));
	  D_RETURN_(-1);
	}
    }
      
  D_RETURN_(0);
  nif = NULL;
}


int
norm_ip_apply_normalizations(NormPacket *p, Interface *nif)
{
  struct ip *iphdr;

  D_ENTER;
  iphdr = (struct ip*)p->network_data;

  /* For incoming packets ... */
  if (!norm_misc_is_trusted_address(iphdr->ip_src))
    {
      /* Optionally clear DF flag and fragment offset --
	 at this point we should have already reassembled
	 the packet, so the fragment offset and MF should be zero.
      */
      if (options.ip_clear_df_and_offset)
	{
	  if (iphdr->ip_off != 0)
	    {
	      norm_stats_increment_ip_norms();
	      print_ip(NORM_LOGLEVEL, ("Clearing DF flag and fragment offset.\n"));
	      iphdr->ip_off = 0;
	      p->network_data_recalc_checksum = TRUE;
	    }
	}

      /* Restore TTL if desiread. */
      if (options.ip_check_ttl)
	{
	  if (iphdr->ip_ttl != options.ip_ttl_value)
	    {
	      norm_stats_increment_ip_norms();
	      print_ip(NORM_LOGLEVEL, ("Setting TTL: %i->%i\n", iphdr->ip_ttl, options.ip_ttl_value));
	      iphdr->ip_ttl = (char)options.ip_ttl_value;
	      p->network_data_recalc_checksum = TRUE;
	    }
	}

    }

  D_RETURN_(0);
  nif = NULL;
}


/** Hand packet off to protocol specific handlers
*/
int
norm_ip_process_protocol(NormPacket *p, Interface *nif)
{
  u_int16_t  off;
  struct ip *iphdr;
  int        result = 0;

  D_ENTER;

  iphdr = (struct ip*) p->network_data;
  p->transport_data =  p->network_data + iphdr->ip_hl*4; 

  off = ntohs(iphdr->ip_off);

  if (((off & IP_OFFMASK) != 0) || (off & IP_MF))
    {
      D_RETURN_(0);
    }

  switch(iphdr->ip_p)
    {
    case IPPROTO_ICMP:
      result = norm_icmp_handler(p, nif);
      break;
    case IPPROTO_IGMP:
      /* result = handle_igmp(data, len, nif); */
      break;
    case IPPROTO_IPIP:
      /* IP in IP - perhaps an Mbone tunnel? */
      break;
    case IPPROTO_TCP:
      result = norm_tcp_handler(p, nif);
      break;
    case IPPROTO_UDP:
      result = norm_udp_handler(p, nif);
      break;
    case IPPROTO_RSVP:
      break;
    case IPPROTO_IGRP:
      break;
    case IPPROTO_PGM:
      break;
    default:
      print_ip(INFO_LOGLEVEL, ("unexpected protocol number: %d\n", iphdr->ip_p));
    }

  D_RETURN_(result);
}


/*
 * Checksum routine for Internet Protocol family headers (C Version)
 */
u_short
norm_ip_in_cksum(register u_short *addr, register int len, unsigned int preadd)
{
  register int nleft = len;
  register u_short *w = addr;
  register u_short answer;
  register int sum = 0;

  D_ENTER;

  sum += preadd;

  /*
   *  Our algorithm is simple, using a 32 bit accumulator (sum),
   *  we add sequential 16 bit words to it, and at the end, fold
   *  back all the carry bits from the top 16 bits into the lower
   *  16 bits.
   */
  while (nleft > 1) {
    sum += *w++;
    nleft -= 2;
  }

  /* mop up an odd byte, if necessary */
  if (nleft == 1)
    sum += *(u_char *)w;

  /*
   * add back carry outs from top 16 bits to low 16 bits
   */
  sum = (sum >> 16) + (sum & 0xffff); /* add hi 16 to low 16 */
  sum += (sum >> 16);                 /* add carry */
  answer = ~sum;                      /* truncate to 16 bits */

  D_RETURN_(answer);
}

#define ADDCARRY(x)  (x > 65535 ? x -= 65535 : x)
#define REDUCE {l_util.l = sum; sum = l_util.s[0] + l_util.s[1]; ADDCARRY(sum); }

int
norm_in_cksum2(unsigned char *m, register int len, unsigned int preadd)
{
  register u_short *w;
  register int sum = 0;
  register int mlen = 0;
  int byte_swapped = 0;
  
  union {
    char    c[2];
    u_short s;
  } s_util;
  union {
    u_short s[2];
    long    l;
  } l_util;
  
  D_ENTER;
  sum += preadd;

  mlen=len;
  w=(u_short*)m;
  /*
   * Force to even boundary.
   */
  if ((1 & (int) w) && (mlen > 0)) {
    REDUCE;
    sum <<= 8;
    s_util.c[0] = *(u_char *)w;
    w = (u_short *)((char *)w + 1);
    mlen--;
    byte_swapped = 1;
  }
  /*
   * Unroll the loop to make overhead from
   * branches &c small.
   */
  while ((mlen -= 32) >= 0) {
    sum += w[0]; sum += w[1]; sum += w[2]; sum += w[3];
    sum += w[4]; sum += w[5]; sum += w[6]; sum += w[7];
    sum += w[8]; sum += w[9]; sum += w[10]; sum += w[11];
    sum += w[12]; sum += w[13]; sum += w[14]; sum += w[15];
    w += 16;
  }
  mlen += 32;
  while ((mlen -= 8) >= 0) {
    sum += w[0]; sum += w[1]; sum += w[2]; sum += w[3];
    w += 4;
  }
  mlen += 8;
  if (mlen == 0 && byte_swapped == 0)
    goto cksum_done;
  REDUCE;
  while ((mlen -= 2) >= 0) {
    sum += *w++;
  }
  if (byte_swapped) {
    REDUCE;
    sum <<= 8;
    byte_swapped = 0;
    if (mlen == -1) {
      s_util.c[1] = *(char *)w;
      sum += s_util.s;
      mlen = 0;
    } else
      mlen = -1;
  } else if (mlen == -1)
    s_util.c[0] = *(char *)w;
 cksum_done:
  if (len)
    printf("cksum: out of data\n");
  if (mlen == -1) {
    /* The last mbuf has odd # of bytes. Follow the
       standard (the odd byte may be shifted left by 8 bits
       or not as determined by endian-ness of the machine) */
    s_util.c[1] = 0;
    sum += s_util.s;
  }
  REDUCE;
  D_RETURN_(~sum & 0xffff);
}


int
norm_ip_checksum_valid(struct ip *iphdr)
{
  unsigned short newsum, origsum;

  D_ENTER;

  origsum=iphdr->ip_sum;
  iphdr->ip_sum=0;

  newsum = norm_ip_in_cksum((u_short *) iphdr, iphdr->ip_hl<<2, 0);

  D(NOISE_LOGLEVEL, ("Checksum in header: %i, calculated: %i\n",
		     ntohs(origsum), ntohs(newsum)));

  if (newsum == origsum) {
    iphdr->ip_sum=origsum;
    D_RETURN_(1);
  }
  iphdr->ip_sum=origsum;

  D_RETURN_(0);
}


