#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <stdlib.h>
#include <stdio.h>

#include <norm.h>
#include <norm_list.h>
#include <norm_debug.h>



NormList *
norm_list_new(void *data)
{
  NormList *l;

  D_ENTER;

  l = NEW(NormList);
  l->data = data;
  l->next = NULL;
  l->prev = NULL;
  
  D_RETURN_(l);
}


int
norm_list_free(NormList *l, NormFunc free_func)
{
  int       length;
  NormList *lnext;

  D_ENTER;

  if (!l)
    D_RETURN_(0);

  length = 0;

  while (l)
    {
      length++;
      lnext = l->next;
      free_func(l->data);
      free(l);
      l = lnext;
    }

  D(NOISE_LOGLEVEL, ("Freed list of %i items\n", length));
  D_RETURN_(length);
}


NormList*
norm_list_find(NormList *l, void *data, NormCmpFunc cmp_func)
{
  D_ENTER;

  while (l)
    {
      if (cmp_func(l->data, data))
	{
	  D(NOISE_LOGLEVEL, ("Successful list find.\n"));
	  D_RETURN_(l);
	}
      
      l = l->next;
    }

  D(NOISE_LOGLEVEL, ("Failed list find.\n"));
  D_RETURN_(NULL);
}


NormList*
norm_list_prepend(NormList *l, void* data)
{
  NormList *lnew;

  D_ENTER;
  lnew = norm_list_new(data);
  lnew->next = l;

  if (l)
    l->prev = lnew;

  D(NOISE_LOGLEVEL, ("Prepending item.\n"));
  D_RETURN_(lnew);
}


void
norm_list_append(NormList *l, void* data)
{
  NormList *lnew;
  NormList *ltmp = NULL;

  D_ENTER;
  if (!l)
    return;

  lnew = norm_list_new(data);

  D(NOISE_LOGLEVEL, ("Appending item.\n")); 

  while (l)
    {
      ltmp = l;
      l = l->next;
    }
  
  ltmp->next = lnew;
  lnew->prev = ltmp;

  D_RETURN;
}


void
norm_list_insert_after(NormList *l, void *data)
{
  NormList *lnew, *lnext;

  D_ENTER;
  if (!l)
    D_RETURN;

  lnew = norm_list_new(data);
  lnext = l->next;

  l->next = lnew;
  lnew->next = lnext;
  lnew->prev = l;
  lnext->prev = lnew;
  
  D(NOISE_LOGLEVEL, ("Inserted item.\n")); 
  D_RETURN;
}


NormList *
norm_list_remove(NormList *l, NormList *item, NormFunc free_func)
{
  NormList *prev;
  NormList *next;

  D_ENTER;
  if (!l)
    D_RETURN_(NULL);

  prev = item->prev;
  next = item->next;

  free_func(item->data);
  free(item);

  /* first item */
  if (!prev)
    {
      if (!next)
	{
	  D_RETURN_(NULL);
	}
      else
	{
	  next->prev = NULL;
	  D_RETURN_(next);
	}
    }

  /* last item */
  if (!next)
    {
      if (!prev)
	{
	  D_RETURN_(NULL);
	}
      else
	{
	  prev->next = NULL;
	  D_RETURN_(l);
	}      
    }

  /* middle item */
  prev->next = next;
  next->prev = prev;

  D_RETURN_(l);
}


NormList *
norm_list_remove_last(NormList *l, NormFunc free_func)
{
  NormList *ll;

  D_ENTER;

  if (!l)
    return (NULL);

  for (ll = l; l; l = l->next)
    ll = l;

  D_RETURN_(norm_list_remove(l, ll, free_func));
}

NormList *
norm_list_move_to_front(NormList *l, NormList *item)
{
  NormList *prev;
  NormList *next;

  D_ENTER;
  if (!l || !item)
    D_RETURN_(NULL);

  prev = item->prev;
  next = item->next;

  /* first item already */
  if (!prev)
    {
      D_RETURN_(l);
    }

  /* last item */
  if (!next)
    {
      prev->next = NULL;
      item->prev = NULL;
      item->next = l;
      l->prev = item;

      D_RETURN_(item);
    }      

  /* middle item */
  prev->next = next;
  next->prev = prev;

  item->next = l;
  item->prev = NULL;
  l->prev = item;

  D_RETURN_(item);
}
